/* Autor: Sergio Blanco Cuaresma
 * Data: Agost 2002
 * Descripci: Graf per optimitzar l'estructura d'un servidor web. Afegeix
 * pgines web amb connexions entre elles i com valors el temps de carrega (ha de ser inferior a 99999).
 * A ms realitza consultes sobre aquestes dades.
 * Aquest TAD no te limit de pgines web.
 */

public class Graf {
  private Llista llistat;

  Graf()  throws NoMemoria, PDIdreta{
    llistat = new Llista();
  }

  // Imprimeix les pagines web guardades
  public void imprimir() throws PDIdreta{
    llistat.imprimir_cadenes();
  }

  // Busca si existeix un element a la llista amb el mateix nom, si el troba retorna el
  // contingut (InfoLlista) sino null. A ms, la llista estara situada en aquest element.
  private InfoLlista buscar(String s) throws NoMemoria, PDIdreta{
    if (llistat.buida()){
      return null;  // Error: no existeix
    }
    llistat.principi(); // Cerca
    while ((!llistat.finalLlista()) && (!s.equals( ((InfoLlista)llistat.actual()).nom()) )){
      llistat.avanzar();
    }
    if (!llistat.finalLlista()) {
      return (InfoLlista)llistat.actual();
    } else {
      return null;  // Error: no existeix
    }
  }

  // Busca un element a la llista amb el mateix nom, si el troba retorna la
  // posicio que ocupa dins de la llista. til per les operacions de consulta.
  private int buscar_posicio(String s) throws NoMemoria, PDIdreta, NoExisteix{
    if (llistat.buida()){
      throw new NoExisteix();
    }
    llistat.principi(); // Cerca
    while ((!llistat.finalLlista()) && (!s.equals( ((InfoLlista)llistat.actual()).nom()) )){
      llistat.avanzar();
    }
    if (!llistat.finalLlista()) {
      return ((InfoLlista)llistat.actual()).posicio();
    } else {
      throw new NoExisteix();
    }
  }

  // Busca l'element en la posici indicada i retorna el seu nom. til per les operacions de consulta.
  private String buscar_nom(int pos) throws NoMemoria, PDIdreta, NoExisteix{
    if (llistat.buida()){
      throw new NoExisteix();
    }
    llistat.principi();
    try {
      llistat.avanzar(pos);   // Ens situem a la posici
    } catch (PDIdreta e) {
      throw new NoExisteix();
    }
    return ((InfoLlista)llistat.actual()).nom();
  }


  // Crea l'origen, el dest i/o la connexi entre ells en cas de no existir. Si ja
  // existia la connexi canvia la seva etiqueta.
  public void afegir_link(String c1, String c2, int n) throws NoMemoria, PDIdreta, TempsNegatiu, NodesIguals{
    InfoLlista i1, i2, nouInfo;
    NodeGraf tmp, nouGraf;


    if (c1.equals(c2)) { // No permitim enllaos amb el mateix, ni temps negatius
      throw new NodesIguals();
    }
    if ((n<0) || (n>=99999)){   // Temps negatio o n superior al valor que representa l'infinit
      throw new TempsNegatiu();
    }
    
    i1 = buscar(c1);  // Busquem node origen
    if (i1 == null) {   // No existeix origen, per tant tampoc connexi
      nouGraf = new NodeGraf(n, null, null, null, null);
      nouInfo = new InfoLlista(c1, nouGraf, null);
      nouGraf.vx(nouInfo);
      nouInfo.posicio(llistat.elements());
      llistat.insertar(nouInfo);
      i2 = buscar(c2);  // Busquem node dest
      if (i2 == null) {   // No existeix dest
        nouInfo = new InfoLlista(c2, null, nouGraf);
        nouGraf.vy(nouInfo);
        nouInfo.posicio(llistat.elements());
        llistat.insertar(nouInfo);
        return; // Creada connexi i dest
      } else {  // Existeix dest
        tmp = i1.fila();
        if (tmp==null){
          i1.fila(nouGraf);
          return; // Creada connexi
        }
        while (tmp.seg_fila()!=null){
          tmp = tmp.seg_fila();
        }
        // tmp conte l'ultim element de la llista del dest
        tmp.seg_fila(nouGraf);  // Afegim a la llista del dest
        return; // Creada connexi
      }
    } else {    // Si existeix origen, podria haver-hi connexi
      tmp = i1.columna(); 
      if (tmp!=null){
        while (tmp.seg_columna()!=null){
          if ( c2.equals(((InfoLlista)tmp.vy()).nom()) ) {    // Existeix la connexi
            tmp.etiqueta(n);
            return;
          }
          tmp = tmp.seg_columna();
        }
        if ( c2.equals(((InfoLlista)tmp.vy()).nom()) ) {    // Existeix la connexi
          tmp.etiqueta(n);
          return;
        }
      }
      // Existeix origen, pero no existeix connexi
      i2 = buscar(c2);
      if (i2 == null) {   // No existeix dest
        nouGraf = new NodeGraf(n, null, null, i1, null);
        nouInfo = new InfoLlista(c2, null, nouGraf);
        nouGraf.vy(nouInfo);
        if (tmp!=null){
          tmp.seg_columna(nouGraf);   // Afegim a la llista del origen
        } else {
          i1.columna(nouGraf);  // Afegim a la llista del origen
        }
        nouInfo.posicio(llistat.elements());
        llistat.insertar(nouInfo);
        return;  // Creat dest i connexi
      } else {  // Existeix dest
        nouGraf = new NodeGraf(n, null, null, i1, i2);
        if (tmp!=null){
          tmp.seg_columna(nouGraf);   // Afegim a la llista del origen
        } else {
          i1.columna(nouGraf);    // Afegim a la llista del origen
        }
        tmp = i2.fila();
        if (tmp!=null){
          while (tmp.seg_fila()!=null){
            tmp = tmp.seg_fila();
          }
          // tmp conte l'ultim element de la llista del dest
          tmp.seg_fila(nouGraf);  // Afegim a la llista del dest
        } else {
          i2.fila(nouGraf);
        }
        return; // Creada connexi
      }
    }
  }

  // Retorna el temps mnim necessari per passar d'una plana c1 a una altra c2
  // Algorisme de Dijkstra
  public int temps_minim_connexio(String c1, String c2) throws NoMemoria, PDIdreta, NoConnexio, NoExisteix{
    TaulaDijkstra taula;
    InfoLlista info;
    NodeGraf tmp;
    int minim, desti;
  
    // Si no existeixen c1 i/o c2, saltara una excepci al buscar_posicio
     
    taula = new TaulaDijkstra(llistat.elements());
    taula.valor(buscar_posicio(c1), 0);  // La posici del origen ha d'estar a 0

    desti = buscar_posicio(c2);

    while (!taula.fi(desti)) {  // Mentre que el calcul per arribar al dest no s'hagui finalitzat

      llistat.principi();
      minim = taula.minim();
      llistat.avanzar(minim);
      info = (InfoLlista)llistat.actual();  // Agafem el que te valor mnim a la taula
      taula.fi(minim, true);  // Ja hem trobat la forma ms ptima d'arribar a aquest
      tmp = info.columna();   // Ens preparem per recorre les seves connexions
      while (tmp!=null){  // Per cada connexi
        // Si aquesta connexi no es troba ja calculada (encara esta pendent) i el seu valor es ms gran que el valor
        // per arribar al origen actual mes l'etiqueta de la connexi llavors...
        if ((!taula.fi( ((InfoLlista)(tmp.vy())).posicio())) && (taula.valor( ((InfoLlista)(tmp.vy())).posicio() ) > taula.valor(info.posicio()) + tmp.etiqueta() )){
          // Actualitzem posici amb nou valor
          taula.valor( ((InfoLlista)(tmp.vy())).posicio(), taula.valor(info.posicio()) + tmp.etiqueta() );
        }
        tmp = tmp.seg_columna();
        
      }
    }
      
    if (taula.valor(desti) == 99999) {  // No hi ha connexio (99999 = infinit)
      throw new NoConnexio();
    }
    return taula.valor(desti);
  }

  // Retorna la seqncia de planes visitades en el pas d'una plana a una altra seguint
  // el cam ptim (respecte al temps total de crrega)
  // Algorisme de Dijkstra amb histric
  public Llista cami_optim(String c1, String c2) throws NoMemoria, PDIdreta, NoConnexio, NoExisteix{
    TaulaDijkstra taula;
    InfoLlista info;
    NodeGraf tmp;
    int minim, desti, origen;
    int[] vector_cami;
    Llista cami=new Llista();

    vector_cami=new int[llistat.elements()];
    // Si no existeixen c1 i/o c2, saltara una excepci al buscar_posicio

    taula = new TaulaDijkstra(llistat.elements());
    origen=buscar_posicio(c1);
    taula.valor(origen, 0);  // La posici del origen ha d'estar a 0

    desti = buscar_posicio(c2);

    while (!taula.fi(desti)) {  // Mentre que el calcul per arribar al dest no s'hagui finalitzat

      llistat.principi();
      minim = taula.minim();
      llistat.avanzar(minim);
      info = (InfoLlista)llistat.actual();  // Agafem el que te valor mnim a la taula
      taula.fi(minim, true);  // Ja hem trobat la forma ms ptima d'arribar a aquest
      tmp = info.columna();   // Ens preparem per recorre les seves connexions
      while (tmp!=null){  // Per cada connexi
        // Si aquesta connexi no es troba ja calculada (encara esta pendent) i el seu valor es ms gran que el valor
        // per arribar al origen actual mes l'etiqueta de la connexi llavors...
        if ((!taula.fi( ((InfoLlista)(tmp.vy())).posicio())) && (taula.valor( ((InfoLlista)(tmp.vy())).posicio() ) > taula.valor(info.posicio()) + tmp.etiqueta() )){
          // Actualitzem posici amb nou valor
          taula.valor( ((InfoLlista)(tmp.vy())).posicio(), taula.valor(info.posicio()) + tmp.etiqueta() );
          // Guardem la posici del node pel que s'ha de passar per arribar aqui
          vector_cami[((InfoLlista)(tmp.vy())).posicio()] = info.posicio();
        }
        tmp = tmp.seg_columna();

      }
    }
    
    if (taula.valor(desti) == 99999) {  // No hi ha connexio (99999 = infinit)
      throw new NoConnexio();
    }
    // Reconstruim cami a partir del histric vector_cami
    construir_cami(origen, desti, vector_cami, cami);

    return cami;
  }

  // Reconstrueix el cam que s'ha de seguir per arribar a un dest determinat basant-se en un vector histric
  // i desprs inserta el nom del dest
  private void construir_cami(int origen, int desti, int[] vector_cami, Llista llista) throws NoMemoria, PDIdreta, NoExisteix{
    construir_cami_re(origen, desti, vector_cami, llista);
    llista.insertar(buscar_nom(desti));
  }
  
  // Reconstrueix el cam que s'ha de seguir per arribar a un dest determinat basant-se en un vector histric
  // de forma recursiva. Es deixa per insertar el nom del dest.
  private void construir_cami_re(int origen, int desti, int[] vector_cami, Llista llista) throws NoMemoria, PDIdreta, NoExisteix{

    if (vector_cami[desti] != origen) {
      construir_cami_re(origen, vector_cami[desti], vector_cami, llista);
      llista.insertar(buscar_nom(vector_cami[desti]));
    } else {
      llista.insertar(buscar_nom(vector_cami[desti]));
    }
  }

  // Comptabilitza el nombre d'enllaos que haurem de travessar per anar d'una plana c1 a una altra c2
  // seguint el cam ptim (respecte al nombre de nodes visitats)
  // Algorisme de Dijkstra ficant totes les etiquetes amb valor 1
  public int nombre_minim_links(String c1, String c2) throws NoMemoria, PDIdreta, NoConnexio, NoExisteix{
    TaulaDijkstra taula;
    InfoLlista info;
    NodeGraf tmp;
    int minim, desti;

    // Si no existeixen c1 i/o c2, saltara una excepci al buscar_posicio

    taula = new TaulaDijkstra(llistat.elements());
    taula.valor(buscar_posicio(c1), 0);  // La posici del origen ha d'estar a 0

    desti = buscar_posicio(c2);

    while (!taula.fi(desti)) {  // Mentre que el calcul per arribar al dest no s'hagui finalitzat

      llistat.principi();
      minim = taula.minim();
      llistat.avanzar(minim);
      info = (InfoLlista)llistat.actual();  // Agafem el que te valor mnim a la taula
      taula.fi(minim, true);  // Ja hem trobat la forma ms ptima d'arribar a aquest
      tmp = info.columna();   // Ens preparem per recorre les seves connexions
      while (tmp!=null){  // Per cada connexi - ATENCI: Considerem totes les etiquetes = 1
        // Si aquesta connexi no es troba ja calculada (encara esta pendent) i el seu valor es ms gran que el valor
        // per arribar al origen actual mes l'etiqueta de la connexi (valor 1 per totes) llavors...
        if ((!taula.fi( ((InfoLlista)(tmp.vy())).posicio())) && (taula.valor( ((InfoLlista)(tmp.vy())).posicio() ) > taula.valor(info.posicio()) + 1 )){
          // Actualitzem posici amb nou valor
          taula.valor( ((InfoLlista)(tmp.vy())).posicio(), taula.valor(info.posicio()) + 1 );
        }
        tmp = tmp.seg_columna();

      }
    }

    if (taula.valor(desti) == 99999) {  // No hi ha connexio (99999 = infinit)
      throw new NoConnexio();
    }
    return taula.valor(desti);
  }


}
      

